"""Tests for motion planning module."""

import pytest

from opentrons.hardware_control.types import CriticalPoint as CP
from opentrons.motion_planning import (
    ArcOutOfBoundsError,
    DestinationOutOfBoundsError,
    MoveType,
    Waypoint,
    get_gripper_labware_movement_waypoints,
    get_waypoints,
)
from opentrons.motion_planning.types import GripperMovementWaypointsWithJawStatus
from opentrons.types import Point


def test_get_waypoints_direct() -> None:
    """It should move directly to points."""
    result = get_waypoints(
        origin=Point(1, 2, 3),
        dest=Point(1, 2, 4),
        move_type=MoveType.DIRECT,
        min_travel_z=0,
        max_travel_z=100,
    )

    assert result == [Waypoint(Point(1, 2, 4))]


def test_get_waypoints_in_labware_arc() -> None:
    """It should use min_travel_z with well_z_margin to path in-labware arc."""
    result = get_waypoints(
        origin=Point(1, 1, 3),
        dest=Point(2, 2, 3),
        move_type=MoveType.IN_LABWARE_ARC,
        min_travel_z=5,
        max_travel_z=100,
    )

    assert result == [
        Waypoint(Point(1, 1, 10)),
        Waypoint(Point(2, 2, 10)),
        Waypoint(Point(2, 2, 3)),
    ]


def test_get_waypoints_in_labware_arc_with_high_dest() -> None:
    """It should favor dest height over travel z if point is higher."""
    result = get_waypoints(
        origin=Point(1, 1, 10),
        dest=Point(2, 2, 11),
        move_type=MoveType.IN_LABWARE_ARC,
        # min_travel_z lower than to and from points
        min_travel_z=3,
        max_travel_z=100,
    )

    assert result == [
        Waypoint(Point(1, 1, 11)),
        Waypoint(Point(2, 2, 11)),
    ]


def test_get_waypoints_in_labware_arc_with_high_origin() -> None:
    """It should favor origin height over travel z if point is higher."""
    result = get_waypoints(
        origin=Point(1, 1, 11),
        dest=Point(2, 2, 10),
        move_type=MoveType.IN_LABWARE_ARC,
        # min_travel_z lower than to and from points
        min_travel_z=3,
        max_travel_z=100,
    )

    assert result == [
        Waypoint(Point(2, 2, 11)),
        Waypoint(Point(2, 2, 10)),
    ]


def test_get_waypoints_in_labware_arc_with_extra_high_origin() -> None:
    """It should favor origin height over max travel z if higher."""
    result = get_waypoints(
        origin=Point(1, 1, 11),
        dest=Point(2, 2, 5),
        move_type=MoveType.IN_LABWARE_ARC,
        min_travel_z=6,
        # max_travel_z lower than starting point
        max_travel_z=10,
    )

    assert result == [
        Waypoint(Point(2, 2, 11)),
        Waypoint(Point(2, 2, 5)),
    ]


def test_get_waypoints_general_arc() -> None:
    """It should use the safe_z with lw_z_margin to path a general arc."""
    result = get_waypoints(
        origin=Point(1, 1, 3),
        dest=Point(2, 2, 3),
        move_type=MoveType.GENERAL_ARC,
        min_travel_z=5,
        max_travel_z=100,
    )

    assert result == [
        Waypoint(Point(1, 1, 15)),
        Waypoint(Point(2, 2, 15)),
        Waypoint(Point(2, 2, 3)),
    ]


def test_get_waypoints_general_arc_with_high_dest() -> None:
    """It should favor dest height over travel z if point is higher."""
    result = get_waypoints(
        origin=Point(1, 1, 14),
        dest=Point(2, 2, 15),
        move_type=MoveType.IN_LABWARE_ARC,
        # min_travel_z lower than to and from points
        min_travel_z=3,
        max_travel_z=100,
    )

    assert result == [
        Waypoint(Point(1, 1, 15)),
        Waypoint(Point(2, 2, 15)),
    ]


def test_get_waypoints_general_arc_with_high_origin() -> None:
    """It should favor origin height over travel z if point is higher."""
    result = get_waypoints(
        origin=Point(1, 1, 15),
        dest=Point(2, 2, 14),
        move_type=MoveType.GENERAL_ARC,
        # min_travel_z lower than to and from points
        min_travel_z=3,
        max_travel_z=100,
    )

    assert result == [
        Waypoint(Point(2, 2, 15)),
        Waypoint(Point(2, 2, 14)),
    ]


def test_get_waypoints_with_extra_waypoints() -> None:
    """It should allow extra XY waypoints."""
    result = get_waypoints(
        origin=Point(1, 1, 3),
        dest=Point(2, 2, 3),
        move_type=MoveType.GENERAL_ARC,
        min_travel_z=5,
        xy_waypoints=[(1, 1.5), (1.5, 1)],
        max_travel_z=100,
    )

    assert result == [
        Waypoint(Point(1, 1, 15)),
        Waypoint(Point(1, 1.5, 15)),
        Waypoint(Point(1.5, 1, 15)),
        Waypoint(Point(2, 2, 15)),
        Waypoint(Point(2, 2, 3)),
    ]


def test_get_waypoints_with_critical_points() -> None:
    """It should attach critical points to origin and destination points."""
    result = get_waypoints(
        origin=Point(1, 1, 3),
        dest=Point(2, 2, 3),
        origin_cp=CP.XY_CENTER,
        dest_cp=CP.FRONT_NOZZLE,
        move_type=MoveType.GENERAL_ARC,
        min_travel_z=5,
        max_travel_z=100,
        xy_waypoints=[(1, 1.5), (1.5, 1)],
    )

    assert result == [
        Waypoint(Point(1, 1, 15), CP.XY_CENTER),
        Waypoint(Point(1, 1.5, 15), CP.FRONT_NOZZLE),
        Waypoint(Point(1.5, 1, 15), CP.FRONT_NOZZLE),
        Waypoint(Point(2, 2, 15), CP.FRONT_NOZZLE),
        Waypoint(Point(2, 2, 3), CP.FRONT_NOZZLE),
    ]


def test_get_waypoints_transitions_cp_with_high_dest() -> None:
    """It should ensure critical_points are blended when dest z is travel z."""
    result = get_waypoints(
        origin=Point(1, 1, 14),
        dest=Point(2, 2, 15),
        dest_cp=CP.XY_CENTER,
        move_type=MoveType.IN_LABWARE_ARC,
        # min_travel_z lower than to and from points
        min_travel_z=3,
        max_travel_z=100,
    )

    assert result == [
        Waypoint(Point(1, 1, 15)),
        Waypoint(Point(2, 2, 15), CP.XY_CENTER),
    ]


def test_get_waypoints_clamps_to_max_travel_z() -> None:
    """It should clamp the travel Z to max_travel_z if it has the clearance."""
    result = get_waypoints(
        origin=Point(1, 1, 3),
        dest=Point(2, 2, 3),
        move_type=MoveType.GENERAL_ARC,
        min_travel_z=5,
        max_travel_z=6,
    )

    assert result == [
        Waypoint(Point(1, 1, 6)),
        Waypoint(Point(2, 2, 6)),
        Waypoint(Point(2, 2, 3)),
    ]


def test_get_waypoints_raises_if_no_max_travel_z_clearance() -> None:
    """Raises ArcOutOfBoundsError if max_travel_z doesn't have clearance."""
    with pytest.raises(ArcOutOfBoundsError):
        get_waypoints(
            origin=Point(1, 1, 3),
            dest=Point(2, 2, 3),
            move_type=MoveType.GENERAL_ARC,
            min_travel_z=5,
            max_travel_z=5.9,
        )


def test_get_waypoints_raises_if_dest_has_no_clearance() -> None:
    """Raises DestinationOutOfBoundsError if dest z doesn't have clearance."""
    with pytest.raises(DestinationOutOfBoundsError):
        get_waypoints(
            origin=Point(1, 1, 3),
            dest=Point(2, 2, 5),
            move_type=MoveType.GENERAL_ARC,
            min_travel_z=5,
            max_travel_z=5.9,
        )


def test_get_waypoints_direct_moves_ignore_clearance_requirements() -> None:
    """It ignores clearance requirements for direct moves."""
    result = get_waypoints(
        origin=Point(1, 1, 3),
        dest=Point(1, 1, 5.5),
        move_type=MoveType.DIRECT,
        min_travel_z=5,
        max_travel_z=5.9,
    )

    assert result == [Waypoint(Point(1, 1, 5.5))]


def test_get_gripper_labware_movement_waypoints() -> None:
    """It should get the correct waypoints for gripper movement."""
    result = get_gripper_labware_movement_waypoints(
        from_labware_center=Point(101, 102, 119.5),
        to_labware_center=Point(201, 202, 219.5),
        gripper_home_z=999,
        post_drop_slide_offset=None,
    )
    assert result == [
        # move to above "from" slot
        GripperMovementWaypointsWithJawStatus(Point(101, 102, 999), False, False),
        # with jaw open, move to labware on "from" slot
        GripperMovementWaypointsWithJawStatus(Point(101, 102, 119.5), True, False),
        # grip labware and retract in place
        GripperMovementWaypointsWithJawStatus(Point(101, 102, 999), False, False),
        # with labware gripped, move to above "to" slot
        GripperMovementWaypointsWithJawStatus(Point(201, 202, 999), False, False),
        # with labware gripped, move down to labware drop height on "to" slot
        GripperMovementWaypointsWithJawStatus(Point(201, 202, 219.5), False, False),
        # ungrip labware and retract in place
        GripperMovementWaypointsWithJawStatus(Point(201, 202, 999), True, True),
    ]


def test_get_gripper_labware_movement_waypoint_with_slide() -> None:
    """It should get the correct waypoints for gripper movement."""
    result = get_gripper_labware_movement_waypoints(
        from_labware_center=Point(101, 102, 119.5),
        to_labware_center=Point(201, 202, 219.5),
        gripper_home_z=999,
        post_drop_slide_offset=Point(x=10, y=10, z=1),
    )
    assert result == [
        # move to above "from" slot
        GripperMovementWaypointsWithJawStatus(Point(101, 102, 999), False, False),
        # with jaw open, move to labware on "from" slot
        GripperMovementWaypointsWithJawStatus(Point(101, 102, 119.5), True, False),
        # grip labware and retract in place
        GripperMovementWaypointsWithJawStatus(Point(101, 102, 999), False, False),
        # with labware gripped, move to above "to" slot
        GripperMovementWaypointsWithJawStatus(Point(201, 202, 999), False, False),
        # with labware gripped, move down to labware drop height on "to" slot
        GripperMovementWaypointsWithJawStatus(Point(201, 202, 219.5), False, False),
        # ungrip labware and retract in place
        GripperMovementWaypointsWithJawStatus(Point(201, 202, 999), True, True),
        # slide after ungripping
        GripperMovementWaypointsWithJawStatus(Point(211, 212, 1000), True, False),
    ]



"""Liquid view tests.

DEPRECATED: Testing LiquidClassView independently of LiquidClassStore is no
longer helpful. Try to add new tests to test_liquid_class_state.py, where they can be
tested together, treating LiquidClassState as a private implementation detail.
"""

import pytest

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    LiquidClassSchemaV1,
)

from opentrons.protocol_engine.state.liquid_classes import (
    LiquidClassState,
    LiquidClassView,
)
from opentrons.protocol_engine.types import LiquidClassRecord


@pytest.fixture
def liquid_class_record(
    minimal_liquid_class_def2: LiquidClassSchemaV1,
) -> LiquidClassRecord:
    """An example LiquidClassRecord for tests."""
    pipette_0 = minimal_liquid_class_def2.byPipette[0]
    by_tip_type_0 = pipette_0.byTipType[0]
    return LiquidClassRecord(
        liquidClassName=minimal_liquid_class_def2.liquidClassName,
        pipetteModel=pipette_0.pipetteModel,
        tiprack=by_tip_type_0.tiprack,
        aspirate=by_tip_type_0.aspirate,
        singleDispense=by_tip_type_0.singleDispense,
        multiDispense=by_tip_type_0.multiDispense,
    )


@pytest.fixture
def subject(liquid_class_record: LiquidClassRecord) -> LiquidClassView:
    """The LiquidClassView test subject."""
    state = LiquidClassState(
        liquid_class_record_by_id={"liquid-class-id": liquid_class_record},
        liquid_class_record_to_id={liquid_class_record: "liquid-class-id"},
    )
    return LiquidClassView(state)


def test_get_by_id(
    subject: LiquidClassView, liquid_class_record: LiquidClassRecord
) -> None:
    """Should look up LiquidClassRecord by ID."""
    assert subject.get("liquid-class-id") == liquid_class_record


def test_get_by_liquid_class_record(
    subject: LiquidClassView, liquid_class_record: LiquidClassRecord
) -> None:
    """Should look up existing ID given a LiquidClassRecord."""
    assert (
        subject.get_id_for_liquid_class_record(liquid_class_record) == "liquid-class-id"
    )


def test_get_all(
    subject: LiquidClassView, liquid_class_record: LiquidClassRecord
) -> None:
    """Should get all LiquidClassRecords in the store."""
    assert subject.get_all() == {"liquid-class-id": liquid_class_record}



"""Test Heater Shaker await temperature command implementation."""

from decoy import Decoy

from opentrons.hardware_control.modules import HeaterShaker
from opentrons.protocol_engine.commands import heater_shaker
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.heater_shaker.wait_for_temperature import (
    WaitForTemperatureImpl,
)
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.state.module_substates import (
    HeaterShakerModuleId,
    HeaterShakerModuleSubState,
)
from opentrons.protocol_engine.state.state import StateView


async def test_wait_for_temperature(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
) -> None:
    """It should be able to wait for the module's target temperature."""
    subject = WaitForTemperatureImpl(state_view=state_view, equipment=equipment)

    data = heater_shaker.WaitForTemperatureParams(moduleId="heater-shaker-id")

    hs_module_substate = decoy.mock(cls=HeaterShakerModuleSubState)
    hs_hardware = decoy.mock(cls=HeaterShaker)

    decoy.when(
        state_view.modules.get_heater_shaker_module_substate(
            module_id="heater-shaker-id"
        )
    ).then_return(hs_module_substate)

    decoy.when(hs_module_substate.get_plate_target_temperature()).then_return(123.45)
    decoy.when(hs_module_substate.module_id).then_return(
        HeaterShakerModuleId("heater-shaker-id")
    )

    # Get stubbed hardware module from hs module view
    decoy.when(
        equipment.get_module_hardware_api(HeaterShakerModuleId("heater-shaker-id"))
    ).then_return(hs_hardware)

    result = await subject.execute(data)
    decoy.verify(
        await hs_hardware.await_temperature(awaiting_temperature=123.45), times=1
    )
    assert result == SuccessData(
        public=heater_shaker.WaitForTemperatureResult(),
    )



"""RailLights handler."""

import pytest
from decoy import Decoy

from opentrons.hardware_control import HardwareControlAPI, OT2HardwareControlAPI
from opentrons.protocol_engine.execution.rail_lights import RailLightsHandler


@pytest.fixture
def subject(
    hardware_api: HardwareControlAPI,
) -> RailLightsHandler:
    """Create a RailLightsHandler with its dependencies mocked out."""
    return RailLightsHandler(hardware_api=hardware_api)


@pytest.mark.parametrize("binary", [True, False])
async def test_set_rail_lights(
    decoy: Decoy,
    subject: RailLightsHandler,
    hardware_api: OT2HardwareControlAPI,
    binary: bool,
) -> None:
    """The hardware controller should be called."""
    await subject.set_rail_lights(on=binary)

    decoy.verify(await hardware_api.set_lights(rails=binary), times=1)



"""Tests for am ChildThreadTransport."""

import threading
from asyncio import get_running_loop
from datetime import datetime
from functools import partial
from typing import Any

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import LabwareDefinition2
from opentrons_shared_data.labware.types import LabwareUri

from opentrons.protocol_engine import DeckPoint, ProtocolEngine, commands
from opentrons.protocol_engine.clients.transports import ChildThreadTransport
from opentrons.protocol_engine.errors import ErrorOccurrence, ProtocolCommandFailedError


@pytest.fixture
async def engine(decoy: Decoy) -> ProtocolEngine:
    """Get a stubbed out ProtocolEngine."""
    return decoy.mock(cls=ProtocolEngine)


@pytest.fixture
async def subject(
    engine: ProtocolEngine,
) -> ChildThreadTransport:
    """Get a ChildThreadTransport test subject."""
    return ChildThreadTransport(engine=engine, loop=get_running_loop())


async def test_execute_command(
    decoy: Decoy,
    engine: ProtocolEngine,
    subject: ChildThreadTransport,
) -> None:
    """It should execute a command synchronously in a child thread."""
    cmd_data = commands.MoveToWellParams(
        pipetteId="pipette-id",
        labwareId="labware-id",
        wellName="A1",
    )
    cmd_result = commands.MoveToWellResult(position=DeckPoint(x=1, y=2, z=3))
    cmd_request = commands.MoveToWellCreate(params=cmd_data)

    decoy.when(await engine.add_and_execute_command(request=cmd_request)).then_return(
        commands.MoveToWell(
            id="cmd-id",
            key="cmd-key",
            status=commands.CommandStatus.SUCCEEDED,
            params=cmd_data,
            result=cmd_result,
            createdAt=datetime.now(),
        )
    )

    task = partial(subject.execute_command, request=cmd_request)
    result = await get_running_loop().run_in_executor(None, task)

    assert result == cmd_result


async def test_execute_command_failure(
    decoy: Decoy,
    engine: ProtocolEngine,
    subject: ChildThreadTransport,
) -> None:
    """It should execute a load labware command."""
    cmd_data = commands.MoveToWellParams(
        pipetteId="pipette-id",
        labwareId="labware-id",
        wellName="A1",
    )
    cmd_request = commands.MoveToWellCreate(params=cmd_data)
    error = ErrorOccurrence(
        id="error-id",
        errorType="PrettyBadError",
        createdAt=datetime(year=2021, month=1, day=1),
        detail="Things are not looking good.",
        errorCode="1234",
    )

    decoy.when(await engine.add_and_execute_command(request=cmd_request)).then_return(
        commands.MoveToWell(
            id="cmd-id",
            key="cmd-key",
            params=cmd_data,
            status=commands.CommandStatus.FAILED,
            error=error,
            createdAt=datetime.now(),
        )
    )

    task = partial(subject.execute_command, request=cmd_request)

    with pytest.raises(ProtocolCommandFailedError):
        await get_running_loop().run_in_executor(None, task)


async def test_call_method(
    decoy: Decoy,
    engine: ProtocolEngine,
    subject: ChildThreadTransport,
) -> None:
    """It should call a synchronous method in a thread-safe manner."""
    labware_def = LabwareDefinition2.model_construct(namespace="hello")  # type: ignore[call-arg]
    labware_uri = LabwareUri("hello/world/123")
    calling_thread_id = None

    def _record_calling_thread(*args: Any, **kwargs: Any) -> LabwareUri:
        nonlocal calling_thread_id
        calling_thread_id = threading.current_thread().ident
        return labware_uri

    decoy.when(engine.add_labware_definition(labware_def)).then_do(
        _record_calling_thread
    )

    _act = partial(
        subject.call_method, "add_labware_definition", definition=labware_def
    )

    result = await get_running_loop().run_in_executor(None, _act)
    assert result == labware_uri
    assert calling_thread_id == threading.current_thread().ident



"""Unit tests for the `command_unions` module."""

from opentrons.protocol_engine.commands import (
    AspirateCreate,
    AspirateParams,
    CustomCreate,
    CustomParams,
    command_unions,
)
from opentrons.util.get_union_elements import get_union_elements


def test_create_types_by_params_type() -> None:
    """CREATE_TYPES_BY_PARAMS_TYPE should map *Params classes to *Create classes.

    We won't test every possible pairing here--just enough to make sure it's basically
    working.
    """
    assert command_unions.CREATE_TYPES_BY_PARAMS_TYPE[AspirateParams] is AspirateCreate
    assert command_unions.CREATE_TYPES_BY_PARAMS_TYPE[CustomParams] is CustomCreate


def test_create_types_by_params_type_is_exhaustive() -> None:
    """CREATE_TYPES_BY_PARAMS_TYPE should have one entry per params type."""
    expected_params_types = frozenset(get_union_elements(command_unions.CommandParams))
    mapped_params_types = frozenset(command_unions.CREATE_TYPES_BY_PARAMS_TYPE.keys())
    assert mapped_params_types == expected_params_types



"""Test load pipette commands."""

import pytest
from decoy import Decoy

from opentrons_shared_data.pipette.pipette_definition import AvailableSensorDefinition
from opentrons_shared_data.pipette.types import (
    LiquidClasses as VolumeModes,
)
from opentrons_shared_data.pipette.types import (
    PipetteNameType,
)

import opentrons.protocol_engine.state.update_types as update_types
from ..pipette_fixtures import get_default_nozzle_map
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.configure_for_volume import (
    ConfigureForVolumeImplementation,
    ConfigureForVolumeParams,
    ConfigureForVolumeResult,
)
from opentrons.protocol_engine.execution import (
    EquipmentHandler,
    LoadedConfigureForVolumeData,
)
from opentrons.protocol_engine.resources.pipette_data_provider import (
    LoadedStaticPipetteData,
)
from opentrons.protocol_engine.types import FlowRates
from opentrons.types import Point


@pytest.fixture
def available_sensors() -> AvailableSensorDefinition:
    """Provide a list of sensors."""
    return AvailableSensorDefinition(sensors=["pressure", "capacitive", "environment"])


@pytest.mark.parametrize(
    "data",
    [
        ConfigureForVolumeParams(
            pipetteId="some id",
            volume=1,
        ),
        ConfigureForVolumeParams(
            pipetteId="some id",
            volume=1,
            tipOverlapNotAfterVersion="v3",
        ),
    ],
)
async def test_configure_for_volume_implementation(
    decoy: Decoy,
    equipment: EquipmentHandler,
    data: ConfigureForVolumeParams,
    available_sensors: AvailableSensorDefinition,
) -> None:
    """A ConfigureForVolume command should have an execution implementation."""
    subject = ConfigureForVolumeImplementation(equipment=equipment)

    config = LoadedStaticPipetteData(
        model="some-model",
        display_name="Hello",
        min_volume=0,
        max_volume=251,
        channels=8,
        home_position=123.1,
        nozzle_offset_z=331.0,
        flow_rates=FlowRates(
            default_aspirate={}, default_dispense={}, default_blow_out={}
        ),
        tip_configuration_lookup_table={},
        nominal_tip_overlap={},
        nozzle_map=get_default_nozzle_map(PipetteNameType.P300_MULTI),
        back_left_corner_offset=Point(10, 20, 30),
        front_right_corner_offset=Point(40, 50, 60),
        pipette_lld_settings={},
        plunger_positions={
            "top": 0.0,
            "bottom": 5.0,
            "blow_out": 19.0,
            "drop_tip": 20.0,
        },
        shaft_ul_per_mm=5.0,
        available_sensors=available_sensors,
        volume_mode=VolumeModes.lowVolumeDefault,
        available_volume_modes_min_vol={},
    )

    decoy.when(
        await equipment.configure_for_volume(
            pipette_id="some id",
            volume=1,
            tip_overlap_version=data.tipOverlapNotAfterVersion,
        )
    ).then_return(
        LoadedConfigureForVolumeData(
            pipette_id="pipette-id",
            serial_number="some number",
            volume=1,
            static_config=config,
        )
    )

    result = await subject.execute(data)

    assert result == SuccessData(
        public=ConfigureForVolumeResult(),
        state_update=update_types.StateUpdate(
            pipette_config=update_types.PipetteConfigUpdate(
                pipette_id="pipette-id", serial_number="some number", config=config
            ),
            ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                pipette_id="pipette-id", ready_to_aspirate=False
            ),
        ),
    )



"""Test load labware commands."""

import inspect
from typing import Optional
from unittest.mock import sentinel

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.load_labware import (
    LoadLabwareImplementation,
    LoadLabwareParams,
    LoadLabwareResult,
)
from opentrons.protocol_engine.errors import (
    LabwareIsNotAllowedInLocationError,
    LocationIsOccupiedError,
)
from opentrons.protocol_engine.execution import EquipmentHandler, LoadedLabwareData
from opentrons.protocol_engine.resources import labware_validation
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.update_types import (
    AddressableAreaUsedUpdate,
    LoadedLabwareUpdate,
    StateUpdate,
)
from opentrons.protocol_engine.types import (
    AddressableAreaLocation,
    DeckSlotLocation,
    LoadableLabwareLocation,
    OnAddressableAreaLocationSequenceComponent,
    OnLabwareLocation,
    OnLabwareLocationSequenceComponent,
)
from opentrons.types import DeckSlotName


@pytest.fixture(autouse=True)
def patch_mock_labware_validation(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out labware_validations.py functions."""
    for name, func in inspect.getmembers(labware_validation, inspect.isfunction):
        monkeypatch.setattr(labware_validation, name, decoy.mock(func=func))


@pytest.mark.parametrize("display_name", ["My custom display name", None])
@pytest.mark.parametrize(
    ("location", "expected_addressable_area_name"),
    [
        (DeckSlotLocation(slotName=DeckSlotName.SLOT_3), "3"),
        (AddressableAreaLocation(addressableAreaName="3"), "3"),
    ],
)
async def test_load_labware_on_slot_or_addressable_area(
    decoy: Decoy,
    well_plate_def: LabwareDefinition,
    equipment: EquipmentHandler,
    state_view: StateView,
    display_name: Optional[str],
    location: LoadableLabwareLocation,
    expected_addressable_area_name: str,
) -> None:
    """A LoadLabware command should have an execution implementation."""
    subject = LoadLabwareImplementation(equipment=equipment, state_view=state_view)

    data = LoadLabwareParams(
        location=location,
        loadName="some-load-name",
        namespace="opentrons-test",
        version=1,
        displayName=display_name,
    )
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            sentinel.validated_empty_location
        )
    ).then_return(
        [
            OnAddressableAreaLocationSequenceComponent(
                addressableAreaName=expected_addressable_area_name,
            )
        ]
    )

    decoy.when(state_view.geometry.ensure_location_not_occupied(location)).then_return(
        sentinel.validated_empty_location
    )
    decoy.when(
        await equipment.load_labware(
            location=sentinel.validated_empty_location,
            load_name="some-load-name",
            namespace="opentrons-test",
            version=1,
            labware_id=None,
        )
    ).then_return(
        LoadedLabwareData(
            labware_id="labware-id",
            definition=well_plate_def,
            offsetId=None,
        )
    )

    decoy.when(
        labware_validation.validate_definition_is_labware(well_plate_def)
    ).then_return(True)

    result = await subject.execute(data)

    assert result == SuccessData(
        public=LoadLabwareResult(
            labwareId="labware-id",
            definition=well_plate_def,
            offsetId=None,
            locationSequence=[
                OnAddressableAreaLocationSequenceComponent(
                    addressableAreaName=expected_addressable_area_name,
                )
            ],
        ),
        state_update=StateUpdate(
            loaded_labware=LoadedLabwareUpdate(
                labware_id="labware-id",
                definition=well_plate_def,
                offset_id=None,
                new_location=sentinel.validated_empty_location,
                display_name=display_name,
            ),
            addressable_area_used=AddressableAreaUsedUpdate(
                addressable_area_name=expected_addressable_area_name
            ),
        ),
    )


async def test_load_labware_raises_location_not_allowed(
    decoy: Decoy,
    equipment: EquipmentHandler,
    state_view: StateView,
) -> None:
    """A LoadLabware command should raise if the flex trash definition is not in a valid slot."""
    subject = LoadLabwareImplementation(equipment=equipment, state_view=state_view)

    data = LoadLabwareParams(
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A2),
        loadName="some-load-name",
        namespace="opentrons-test",
        version=1,
        displayName="My custom display name",
    )

    decoy.when(labware_validation.is_flex_trash("some-load-name")).then_return(True)

    with pytest.raises(LabwareIsNotAllowedInLocationError):
        await subject.execute(data)


async def test_load_labware_on_labware(
    decoy: Decoy,
    well_plate_def: LabwareDefinition,
    equipment: EquipmentHandler,
    state_view: StateView,
) -> None:
    """A LoadLabware command should raise if the definition is not validated as a labware."""
    subject = LoadLabwareImplementation(equipment=equipment, state_view=state_view)

    data = LoadLabwareParams(
        location=OnLabwareLocation(labwareId="other-labware-id"),
        loadName="some-load-name",
        namespace="opentrons-test",
        version=1,
        displayName="My custom display name",
    )

    decoy.when(
        state_view.geometry.ensure_location_not_occupied(
            OnLabwareLocation(labwareId="other-labware-id")
        )
    ).then_return(OnLabwareLocation(labwareId="another-labware-id"))
    decoy.when(
        await equipment.load_labware(
            location=OnLabwareLocation(labwareId="another-labware-id"),
            load_name="some-load-name",
            namespace="opentrons-test",
            version=1,
            labware_id=None,
        )
    ).then_return(
        LoadedLabwareData(
            labware_id="labware-id",
            definition=well_plate_def,
            offsetId="labware-offset-id",
        )
    )
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            OnLabwareLocation(labwareId="another-labware-id")
        )
    ).then_return(
        [
            OnLabwareLocationSequenceComponent(
                labwareId="other-labware-id", lidId=None
            ),
            OnAddressableAreaLocationSequenceComponent(addressableAreaName="A3"),
        ]
    )

    decoy.when(
        labware_validation.validate_definition_is_labware(well_plate_def)
    ).then_return(True)

    result = await subject.execute(data)
    assert result == SuccessData(
        public=LoadLabwareResult(
            labwareId="labware-id",
            definition=well_plate_def,
            offsetId="labware-offset-id",
            locationSequence=[
                OnLabwareLocationSequenceComponent(
                    labwareId="other-labware-id", lidId=None
                ),
                OnAddressableAreaLocationSequenceComponent(addressableAreaName="A3"),
            ],
        ),
        state_update=StateUpdate(
            loaded_labware=LoadedLabwareUpdate(
                labware_id="labware-id",
                definition=well_plate_def,
                offset_id="labware-offset-id",
                new_location=OnLabwareLocation(labwareId="another-labware-id"),
                display_name="My custom display name",
            )
        ),
    )

    decoy.verify(
        state_view.labware.raise_if_labware_cannot_be_stacked(
            well_plate_def, "another-labware-id"
        )
    )


async def test_load_labware_raises_if_location_occupied(
    decoy: Decoy,
    well_plate_def: LabwareDefinition,
    equipment: EquipmentHandler,
    state_view: StateView,
) -> None:
    """A LoadLabware command should have an execution implementation."""
    subject = LoadLabwareImplementation(equipment=equipment, state_view=state_view)

    data = LoadLabwareParams(
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
        loadName="some-load-name",
        namespace="opentrons-test",
        version=1,
        displayName="My custom display name",
    )

    decoy.when(
        state_view.geometry.ensure_location_not_occupied(
            DeckSlotLocation(slotName=DeckSlotName.SLOT_3)
        )
    ).then_raise(LocationIsOccupiedError("Get your own spot!"))

    with pytest.raises(LocationIsOccupiedError):
        await subject.execute(data)



"""Tests for the PythonAndLegacyRunner's LegacyCommandMapper."""

import inspect
from datetime import datetime
from typing import cast

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.labware.types import LabwareDefinition2
from opentrons_shared_data.module.types import ModuleDefinitionV3
from opentrons_shared_data.pipette.types import PipetteNameType

from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.hardware_control.modules.types import TemperatureModuleModel
from opentrons.legacy_commands.types import CommandMessage, CommentMessage, PauseMessage
from opentrons.protocol_api.core.legacy.load_info import (
    InstrumentLoadInfo as LegacyInstrumentLoadInfo,
)
from opentrons.protocol_api.core.legacy.load_info import (
    LabwareLoadInfo as LegacyLabwareLoadInfo,
)
from opentrons.protocol_api.core.legacy.load_info import (
    ModuleLoadInfo as LegacyModuleLoadInfo,
)
from opentrons.protocol_engine import (
    ModuleDefinition,
    ModuleModel,
)
from opentrons.protocol_engine import (
    actions as pe_actions,
)
from opentrons.protocol_engine import (
    commands as pe_commands,
)
from opentrons.protocol_engine import (
    types as pe_types,
)
from opentrons.protocol_engine.error_recovery_policy import ErrorRecoveryType
from opentrons.protocol_engine.resources import (
    ModuleDataProvider,
    pipette_data_provider,
)
from opentrons.protocol_engine.resources.pipette_data_provider import (
    LoadedStaticPipetteData,
)
from opentrons.protocol_engine.state.update_types import (
    LoadedLabwareUpdate,
    LoadModuleUpdate,
    LoadPipetteUpdate,
    PipetteConfigUpdate,
    StateUpdate,
)
from opentrons.protocol_runner.legacy_command_mapper import (
    LegacyCommandMapper,
    LegacyContextCommandError,
)
from opentrons.types import DeckSlotName, Mount, MountType


@pytest.fixture
def module_data_provider(decoy: Decoy) -> ModuleDataProvider:
    """Mock module definition fetcher."""
    return decoy.mock(cls=ModuleDataProvider)


@pytest.fixture(autouse=True)
def _use_mock_pipette_data_provider(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock module definition fetcher."""
    for name, func in inspect.getmembers(pipette_data_provider, inspect.isfunction):
        monkeypatch.setattr(pipette_data_provider, name, decoy.mock(func=func))


def test_map_before_command() -> None:
    """It should map a "before" message to a running command."""
    legacy_command: CommentMessage = {
        "$": "before",
        "id": "message-id",
        "name": "command.COMMENT",
        "payload": {"text": "hello world"},
        "error": None,
    }

    subject = LegacyCommandMapper()
    result = subject.map_command(legacy_command)

    assert result == [
        pe_actions.QueueCommandAction(
            command_id="command.COMMENT-0",
            created_at=matchers.IsA(datetime),
            request=pe_commands.CommentCreate(
                key="command.COMMENT-0",
                params=pe_commands.CommentParams(
                    message="hello world",
                ),
            ),
            request_hash=None,
        ),
        pe_actions.RunCommandAction(
            command_id="command.COMMENT-0",
            started_at=matchers.IsA(datetime),
        ),
    ]


def test_map_after_command() -> None:
    """It should map an "after" message to a succeeded command."""
    legacy_command_start: CommentMessage = {
        "$": "before",
        "id": "message-id",
        "name": "command.COMMENT",
        "payload": {"text": "hello world"},
        "error": None,
    }
    legacy_command_end: CommentMessage = {
        "$": "after",
        "id": "message-id",
        "name": "command.COMMENT",
        "payload": {"text": "hello world"},
        "error": None,
    }

    subject = LegacyCommandMapper()

    _ = subject.map_command(legacy_command_start)
    result = subject.map_command(legacy_command_end)

    assert result == [
        pe_actions.SucceedCommandAction(
            command=pe_commands.Comment.model_construct(
                id="command.COMMENT-0",
                key="command.COMMENT-0",
                status=pe_commands.CommandStatus.SUCCEEDED,
                createdAt=matchers.IsA(datetime),
                startedAt=matchers.IsA(datetime),
                completedAt=matchers.IsA(datetime),
                params=pe_commands.CommentParams(
                    message="hello world",
                ),
                result=pe_commands.CommentResult(),
                notes=[],
            ),
        )
    ]


def test_map_after_with_error_command() -> None:
    """It should map an "after" message to a failed command."""
    legacy_command_start: CommentMessage = {
        "$": "before",
        "id": "message-id",
        "name": "command.COMMENT",
        "error": None,
        "payload": {"text": "hello world"},
    }
    legacy_command_end: CommentMessage = {
        "$": "after",
        "id": "message-id",
        "name": "command.COMMENT",
        "error": RuntimeError("oh no"),
        "payload": {"text": "hello world"},
    }
    subject = LegacyCommandMapper()

    _ = subject.map_command(legacy_command_start)
    result = subject.map_command(legacy_command_end)

    assert result == [
        pe_actions.FailCommandAction(
            command_id="command.COMMENT-0",
            running_command=matchers.Anything(),
            error_id=matchers.IsA(str),
            failed_at=matchers.IsA(datetime),
            error=matchers.ErrorMatching(
                LegacyContextCommandError,
                match="oh no",
            ),
            notes=[],
            type=ErrorRecoveryType.FAIL_RUN,
        )
    ]


def test_command_stack() -> None:
    """It should use messages ID and command ordering to create command IDs."""
    legacy_command_1: CommentMessage = {
        "$": "before",
        "id": "message-id-1",
        "name": "command.COMMENT",
        "payload": {"text": "hello"},
        "error": None,
    }
    legacy_command_2: CommentMessage = {
        "$": "before",
        "id": "message-id-2",
        "name": "command.COMMENT",
        "payload": {"text": "goodbye"},
        "error": None,
    }
    legacy_command_3: CommentMessage = {
        "$": "after",
        "id": "message-id-1",
        "name": "command.COMMENT",
        "payload": {"text": "hello"},
        "error": None,
    }
    legacy_command_4: CommentMessage = {
        "$": "after",
        "id": "message-id-2",
        "name": "command.COMMENT",
        "payload": {"text": "goodbye"},
        "error": RuntimeError("oh no"),
    }

    subject = LegacyCommandMapper()
    result = [
        *subject.map_command(legacy_command_1),
        *subject.map_command(legacy_command_2),
        *subject.map_command(legacy_command_3),
        *subject.map_command(legacy_command_4),
    ]

    assert result == [
        pe_actions.QueueCommandAction(
            command_id="command.COMMENT-0",
            created_at=matchers.IsA(datetime),
            request=pe_commands.CommentCreate(
                key="command.COMMENT-0",
                params=pe_commands.CommentParams(
                    message="hello",
                ),
            ),
            request_hash=None,
        ),
        pe_actions.RunCommandAction(
            command_id="command.COMMENT-0", started_at=matchers.IsA(datetime)
        ),
        pe_actions.QueueCommandAction(
            command_id="command.COMMENT-1",
            created_at=matchers.IsA(datetime),
            request=pe_commands.CommentCreate(
                key="command.COMMENT-1",
                params=pe_commands.CommentParams(
                    message="goodbye",
                ),
            ),
            request_hash=None,
        ),
        pe_actions.RunCommandAction(
            command_id="command.COMMENT-1", started_at=matchers.IsA(datetime)
        ),
        pe_actions.SucceedCommandAction(
            command=pe_commands.Comment.model_construct(
                id="command.COMMENT-0",
                key="command.COMMENT-0",
                status=pe_commands.CommandStatus.SUCCEEDED,
                createdAt=matchers.IsA(datetime),
                startedAt=matchers.IsA(datetime),
                completedAt=matchers.IsA(datetime),
                params=pe_commands.CommentParams(
                    message="hello",
                ),
                result=pe_commands.CommentResult(),
                notes=[],
            ),
        ),
        pe_actions.FailCommandAction(
            command_id="command.COMMENT-1",
            running_command=matchers.Anything(),
            error_id=matchers.IsA(str),
            failed_at=matchers.IsA(datetime),
            error=matchers.ErrorMatching(LegacyContextCommandError, "oh no"),
            notes=[],
            type=ErrorRecoveryType.FAIL_RUN,
        ),
    ]


def test_map_labware_load(minimal_labware_def: LabwareDefinition2) -> None:
    """It should correctly map a labware load."""
    input = LegacyLabwareLoadInfo(
        labware_definition=minimal_labware_def,
        labware_namespace="some_namespace",
        labware_load_name="some_load_name",
        labware_version=123,
        deck_slot=DeckSlotName.SLOT_1,
        on_module=False,
        offset_id="labware-offset-id-123",
        labware_display_name="My special labware",
    )

    expected_id_and_key = "commands.LOAD_LABWARE-0"
    expected_params = pe_commands.LoadLabwareParams(
        location=pe_types.DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        namespace="some_namespace",
        loadName="some_load_name",
        version=123,
        displayName="My special labware",
        labwareId=None,
    )
    expected_queue = pe_actions.QueueCommandAction(
        command_id=expected_id_and_key,
        created_at=matchers.IsA(datetime),
        request=pe_commands.LoadLabwareCreate(
            key=expected_id_and_key,
            params=expected_params,
        ),
        request_hash=None,
    )
    expected_run = pe_actions.RunCommandAction(
        command_id=expected_id_and_key,
        started_at=matchers.IsA(datetime),
    )
    expected_succeed = pe_actions.SucceedCommandAction(
        command=pe_commands.LoadLabware.model_construct(
            id=expected_id_and_key,
            key=expected_id_and_key,
            params=expected_params,
            status=pe_commands.CommandStatus.SUCCEEDED,
            createdAt=matchers.IsA(datetime),
            startedAt=matchers.IsA(datetime),
            completedAt=matchers.IsA(datetime),
            result=pe_commands.LoadLabwareResult.model_construct(
                labwareId=matchers.IsA(str),
                # Trusting that the exact fields within in the labware definition
                # get passed through correctly.
                definition=matchers.Anything(),
                offsetId="labware-offset-id-123",
                locationSequence=[
                    pe_types.OnAddressableAreaLocationSequenceComponent(
                        addressableAreaName="1"
                    )
                ],
            ),
            notes=[],
        ),
        state_update=StateUpdate(
            loaded_labware=LoadedLabwareUpdate(
                labware_id="labware-0",
                definition=matchers.Anything(),
                offset_id="labware-offset-id-123",
                new_location=pe_types.DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
                display_name="My special labware",
            )
        ),
    )
    result_queue, result_run, result_succeed = LegacyCommandMapper().map_equipment_load(
        input
    )
    assert result_queue == expected_queue
    assert result_run == expected_run
    assert result_succeed == expected_succeed


def test_map_instrument_load(decoy: Decoy) -> None:
    """It should correctly map an instrument load."""
    pipette_dict = cast(PipetteDict, {"pipette_id": "fizzbuzz"})
    input = LegacyInstrumentLoadInfo(
        instrument_load_name="p1000_single_gen2",
        mount=Mount.LEFT,
        pipette_dict=pipette_dict,
    )
    pipette_config = cast(LoadedStaticPipetteData, {"config": True})

    decoy.when(
        pipette_data_provider.get_pipette_static_config(pipette_dict, "v0"),
    ).then_return(pipette_config)

    expected_id_and_key = "commands.LOAD_PIPETTE-0"
    expected_params = pe_commands.LoadPipetteParams.model_construct(
        pipetteName=PipetteNameType.P1000_SINGLE_GEN2, mount=MountType.LEFT
    )
    expected_queue = pe_actions.QueueCommandAction(
        command_id=expected_id_and_key,
        created_at=matchers.IsA(datetime),
        request=pe_commands.LoadPipetteCreate(
            key=expected_id_and_key, params=expected_params
        ),
        request_hash=None,
    )
    expected_run = pe_actions.RunCommandAction(
        command_id=expected_id_and_key, started_at=matchers.IsA(datetime)
    )
    expected_succeed = pe_actions.SucceedCommandAction(
        command=pe_commands.LoadPipette.model_construct(
            id=expected_id_and_key,
            key=expected_id_and_key,
            status=pe_commands.CommandStatus.SUCCEEDED,
            createdAt=matchers.IsA(datetime),
            startedAt=matchers.IsA(datetime),
            completedAt=matchers.IsA(datetime),
            params=expected_params,
            result=pe_commands.LoadPipetteResult(pipetteId="pipette-0"),
            notes=[],
        ),
        state_update=StateUpdate(
            loaded_pipette=LoadPipetteUpdate(
                pipette_id="pipette-0",
                mount=expected_params.mount,
                pipette_name=expected_params.pipetteName,
                liquid_presence_detection=expected_params.liquidPresenceDetection,
            ),
            pipette_config=PipetteConfigUpdate(
                pipette_id="pipette-0",
                serial_number="fizzbuzz",
                config=pipette_config,
            ),
        ),
    )

    [
        result_queue,
        result_run,
        result_succeed,
    ] = LegacyCommandMapper().map_equipment_load(input)

    assert result_queue == expected_queue
    assert result_run == expected_run
    assert result_succeed == expected_succeed


def test_map_module_load(
    decoy: Decoy,
    minimal_module_def: ModuleDefinitionV3,
    module_data_provider: ModuleDataProvider,
) -> None:
    """It should correctly map a module load."""
    test_definition = ModuleDefinition.model_validate(minimal_module_def)
    input = LegacyModuleLoadInfo(
        requested_model=TemperatureModuleModel.TEMPERATURE_V1,
        loaded_model=TemperatureModuleModel.TEMPERATURE_V2,
        deck_slot=DeckSlotName.SLOT_1,
        configuration="conf",
        module_serial="module-serial",
    )
    decoy.when(
        module_data_provider.get_definition(ModuleModel.TEMPERATURE_MODULE_V2)
    ).then_return(test_definition)

    expected_id_and_key = "commands.LOAD_MODULE-0"
    expected_params = pe_commands.LoadModuleParams.model_construct(
        model=ModuleModel.TEMPERATURE_MODULE_V1,
        location=pe_types.DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        moduleId=matchers.IsA(str),
    )
    expected_queue = pe_actions.QueueCommandAction(
        command_id=expected_id_and_key,
        created_at=matchers.IsA(datetime),
        request=pe_commands.LoadModuleCreate(
            key=expected_id_and_key, params=expected_params
        ),
        request_hash=None,
    )
    expected_run = pe_actions.RunCommandAction(
        command_id=expected_id_and_key, started_at=matchers.IsA(datetime)
    )
    expected_succeed = pe_actions.SucceedCommandAction(
        command=pe_commands.LoadModule.model_construct(
            id=expected_id_and_key,
            key=expected_id_and_key,
            status=pe_commands.CommandStatus.SUCCEEDED,
            createdAt=matchers.IsA(datetime),
            startedAt=matchers.IsA(datetime),
            completedAt=matchers.IsA(datetime),
            params=expected_params,
            result=pe_commands.LoadModuleResult.model_construct(
                moduleId=matchers.IsA(str),
                serialNumber="module-serial",
                model=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            notes=[],
        ),
        state_update=StateUpdate(
            loaded_module=LoadModuleUpdate(
                module_id=matchers.IsA(str),
                definition=test_definition,
                slot_name=DeckSlotName.SLOT_1,
                requested_model=ModuleModel.TEMPERATURE_MODULE_V1,
                serial_number="module-serial",
            )
        ),
    )

    [result_queue, result_run, result_succeed] = LegacyCommandMapper(
        module_data_provider=module_data_provider
    ).map_equipment_load(input)

    assert result_queue == expected_queue
    assert result_run == expected_run
    assert result_succeed == expected_succeed


def test_map_module_labware_load(minimal_labware_def: LabwareDefinition2) -> None:
    """It should correctly map a labware load on module."""
    load_input = LegacyLabwareLoadInfo(
        labware_definition=minimal_labware_def,
        labware_namespace="some_namespace",
        labware_load_name="some_load_name",
        labware_display_name="My very special module labware",
        labware_version=123,
        deck_slot=DeckSlotName.SLOT_1,
        on_module=True,
        offset_id="labware-offset-id-123",
    )

    expected_id_and_key = "commands.LOAD_LABWARE-0"
    expected_params = pe_commands.LoadLabwareParams.model_construct(
        location=pe_types.ModuleLocation(moduleId="module-123"),
        namespace="some_namespace",
        loadName="some_load_name",
        version=123,
        displayName="My very special module labware",
        labwareId=None,
    )
    expected_queue = pe_actions.QueueCommandAction(
        command_id=expected_id_and_key,
        created_at=matchers.IsA(datetime),
        request=pe_commands.LoadLabwareCreate(
            key=expected_id_and_key,
            params=expected_params,
        ),
        request_hash=None,
    )
    expected_run = pe_actions.RunCommandAction(
        command_id="commands.LOAD_LABWARE-0",
        started_at=matchers.IsA(datetime),
    )
    expected_succeed = pe_actions.SucceedCommandAction(
        command=pe_commands.LoadLabware.model_construct(
            id=expected_id_and_key,
            key=expected_id_and_key,
            params=expected_params,
            status=pe_commands.CommandStatus.SUCCEEDED,
            createdAt=matchers.IsA(datetime),
            startedAt=matchers.IsA(datetime),
            completedAt=matchers.IsA(datetime),
            result=pe_commands.LoadLabwareResult.model_construct(
                labwareId=matchers.IsA(str),
                # Trusting that the exact fields within in the labware definition
                # get passed through correctly.
                definition=matchers.Anything(),
                offsetId="labware-offset-id-123",
                locationSequence=[
                    pe_types.OnModuleLocationSequenceComponent(moduleId="module-123"),
                    pe_types.OnAddressableAreaLocationSequenceComponent(
                        addressableAreaName="1"
                    ),
                ],
            ),
            notes=[],
        ),
        state_update=StateUpdate(
            loaded_labware=LoadedLabwareUpdate(
                labware_id="labware-0",
                definition=matchers.Anything(),
                offset_id="labware-offset-id-123",
                new_location=pe_types.ModuleLocation(moduleId="module-123"),
                display_name="My very special module labware",
            )
        ),
    )

    subject = LegacyCommandMapper()
    subject._module_id_by_slot = {DeckSlotName.SLOT_1: "module-123"}
    result_queue, result_run, result_succeed = subject.map_equipment_load(load_input)

    assert result_queue == expected_queue
    assert result_run == expected_run
    assert result_succeed == expected_succeed


def test_map_pause() -> None:
    """It should map an "command.PAUSE" message."""
    legacy_command_start: PauseMessage = {
        "$": "before",
        "id": "message-id",
        "name": "command.PAUSE",
        "payload": {"userMessage": "hello world", "text": "hello world"},
        "error": None,
    }
    legacy_command_end: PauseMessage = {
        "$": "after",
        "id": "message-id",
        "name": "command.PAUSE",
        "payload": {"userMessage": "hello world", "text": "hello world"},
        "error": None,
    }

    subject = LegacyCommandMapper()
    result = [
        *subject.map_command(legacy_command_start),
        *subject.map_command(legacy_command_end),
    ]

    assert result == [
        pe_actions.QueueCommandAction(
            command_id="command.PAUSE-0",
            created_at=matchers.IsA(datetime),
            request=pe_commands.WaitForResumeCreate(
                key="command.PAUSE-0",
                params=pe_commands.WaitForResumeParams(message="hello world"),
            ),
            request_hash=None,
        ),
        pe_actions.RunCommandAction(
            command_id="command.PAUSE-0",
            started_at=matchers.IsA(datetime),
        ),
        pe_actions.SucceedCommandAction(
            command=pe_commands.WaitForResume.model_construct(
                id="command.PAUSE-0",
                key="command.PAUSE-0",
                status=pe_commands.CommandStatus.SUCCEEDED,
                createdAt=matchers.IsA(datetime),
                startedAt=matchers.IsA(datetime),
                completedAt=matchers.IsA(datetime),
                params=pe_commands.WaitForResumeParams(message="hello world"),
                notes=[],
            ),
        ),
        pe_actions.PauseAction(source=pe_actions.PauseSource.PROTOCOL),
    ]


@pytest.mark.parametrize(
    "command_type",
    [
        "command.MIX",
        "command.CONSOLIDATE",
        "command.DISTRIBUTE",
        "command.TRANSFER",
        "command.RETURN_TIP",
        "command.AIR_GAP",
    ],
)
def test_filter_higher_order_commands(command_type: str) -> None:
    """It should filter out higher order commands."""
    legacy_command_start = cast(
        CommandMessage,
        {
            "$": "before",
            "id": "message-id",
            "name": command_type,
            "payload": {"text": "hello world"},
            "error": None,
        },
    )
    legacy_command_end = cast(
        CommandMessage,
        {
            "$": "after",
            "id": "message-id",
            "name": command_type,
            "payload": {"text": "hello world"},
            "error": None,
        },
    )

    subject = LegacyCommandMapper()
    result = [
        *subject.map_command(legacy_command_start),
        *subject.map_command(legacy_command_end),
    ]

    assert result == []



"""Unit tests for the utilities in `update_types`."""

from opentrons.protocol_engine import DeckSlotLocation, ModuleLocation
from opentrons.protocol_engine.state import update_types
from opentrons.types import DeckSlotName


def test_append() -> None:
    """Test `StateUpdate.append()`."""
    state_update = update_types.StateUpdate(
        labware_location=update_types.LabwareLocationUpdate(
            labware_id="test-123",
            new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A1),
            offset_id=None,
        )
    )

    # Populating a new field should leave the original ones unchanged.
    result = state_update.append(
        update_types.StateUpdate(pipette_location=update_types.CLEAR)
    )
    assert result is state_update
    assert state_update.labware_location == update_types.LabwareLocationUpdate(
        labware_id="test-123",
        new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A1),
        offset_id=None,
    )
    assert state_update.pipette_location == update_types.CLEAR

    # Populating a field that's already been populated should overwrite it.
    result = state_update.append(
        update_types.StateUpdate(
            labware_location=update_types.LabwareLocationUpdate(
                labware_id="test-123",
                new_location=ModuleLocation(moduleId="module-123"),
                offset_id=None,
            )
        )
    )
    assert result is state_update
    assert state_update.labware_location == update_types.LabwareLocationUpdate(
        labware_id="test-123",
        new_location=ModuleLocation(moduleId="module-123"),
        offset_id=None,
    )
    assert state_update.pipette_location == update_types.CLEAR


def test_reduce() -> None:
    """Test `StateUpdate.reduce()`."""
    assert update_types.StateUpdate.reduce() == update_types.StateUpdate()

    # It should union all the set fields together.
    assert update_types.StateUpdate.reduce(
        update_types.StateUpdate(
            labware_location=update_types.LabwareLocationUpdate(
                labware_id="test-123",
                new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A1),
                offset_id=None,
            )
        ),
        update_types.StateUpdate(pipette_location=update_types.CLEAR),
    ) == update_types.StateUpdate(
        labware_location=update_types.LabwareLocationUpdate(
            labware_id="test-123",
            new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A1),
            offset_id=None,
        ),
        pipette_location=update_types.CLEAR,
    )

    # When one field appears multiple times, the last write wins.
    assert update_types.StateUpdate.reduce(
        update_types.StateUpdate(
            labware_location=update_types.LabwareLocationUpdate(
                labware_id="test-123",
                new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A1),
                offset_id=None,
            )
        ),
        update_types.StateUpdate(
            labware_location=update_types.LabwareLocationUpdate(
                labware_id="test-123",
                new_location=ModuleLocation(moduleId="module-123"),
                offset_id=None,
            )
        ),
    ) == update_types.StateUpdate(
        labware_location=update_types.LabwareLocationUpdate(
            labware_id="test-123",
            new_location=ModuleLocation(moduleId="module-123"),
            offset_id=None,
        )
    )



"""Tests for Protocol API temperature module contexts."""

import pytest
from decoy import Decoy, matchers

from opentrons.drivers.types import HeaterShakerLabwareLatchStatus
from opentrons.hardware_control.modules import SpeedStatus, TemperatureStatus
from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_api import MAX_SUPPORTED_VERSION, HeaterShakerContext
from opentrons.protocol_api.core.common import HeaterShakerCore, ProtocolCore
from opentrons.protocol_api.core.core_map import LoadedCoreMap
from opentrons.protocol_api.tasks import Task
from opentrons.protocols.api_support.types import APIVersion


@pytest.fixture
def mock_core(decoy: Decoy) -> HeaterShakerCore:
    """Get a mock module implementation core."""
    return decoy.mock(cls=HeaterShakerCore)


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol implementation core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def mock_core_map(decoy: Decoy) -> LoadedCoreMap:
    """Get a mock LoadedCoreMap."""
    return decoy.mock(cls=LoadedCoreMap)


@pytest.fixture
def mock_broker(decoy: Decoy) -> LegacyBroker:
    """Get a mock command message broker."""
    return decoy.mock(cls=LegacyBroker)


@pytest.fixture
def api_version() -> APIVersion:
    """Get an API version to apply to the interface."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def subject(
    api_version: APIVersion,
    mock_core: HeaterShakerCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_broker: LegacyBroker,
) -> HeaterShakerContext:
    """Get a temperature module context with its dependencies mocked out."""
    return HeaterShakerContext(
        core=mock_core,
        protocol_core=mock_protocol_core,
        core_map=mock_core_map,
        broker=mock_broker,
        api_version=api_version,
    )


def test_get_current_temperature(
    decoy: Decoy, mock_core: HeaterShakerCore, subject: HeaterShakerContext
) -> None:
    """It should get the current temperature from the core."""
    decoy.when(mock_core.get_current_temperature()).then_return(42.0)
    result = subject.current_temperature
    assert result == 42.0


def test_get_target_temperature(
    decoy: Decoy, mock_core: HeaterShakerCore, subject: HeaterShakerContext
) -> None:
    """It should get the target temperature from the core."""
    decoy.when(mock_core.get_target_temperature()).then_return(42.0)
    result = subject.target_temperature
    assert result == 42.0


def test_get_current_speed(
    decoy: Decoy, mock_core: HeaterShakerCore, subject: HeaterShakerContext
) -> None:
    """It should get the current temperature from the core."""
    decoy.when(mock_core.get_current_speed()).then_return(321)
    result = subject.current_speed
    assert result == 321


def test_get_target_speed(
    decoy: Decoy, mock_core: HeaterShakerCore, subject: HeaterShakerContext
) -> None:
    """It should get the target temperature from the core."""
    decoy.when(mock_core.get_target_speed()).then_return(321)
    result = subject.target_speed
    assert result == 321


def test_get_temperature_status(
    decoy: Decoy, mock_core: HeaterShakerCore, subject: HeaterShakerContext
) -> None:
    """It should get the temperature status from the core."""
    decoy.when(mock_core.get_temperature_status()).then_return(
        TemperatureStatus.HEATING
    )
    result = subject.temperature_status
    assert result == "heating"


def test_get_speed_status(
    decoy: Decoy, mock_core: HeaterShakerCore, subject: HeaterShakerContext
) -> None:
    """It should get the speed status from the core."""
    decoy.when(mock_core.get_speed_status()).then_return(SpeedStatus.ACCELERATING)
    result = subject.speed_status
    assert result == "speeding up"


def test_get_labware_latch_status(
    decoy: Decoy, mock_core: HeaterShakerCore, subject: HeaterShakerContext
) -> None:
    """It should get the speed status from the core."""
    decoy.when(mock_core.get_labware_latch_status()).then_return(
        HeaterShakerLabwareLatchStatus.OPENING
    )
    result = subject.labware_latch_status
    assert result == "opening"


def test_set_target_temperature(
    decoy: Decoy,
    mock_core: HeaterShakerCore,
    mock_broker: LegacyBroker,
    subject: HeaterShakerContext,
    api_version: APIVersion,
) -> None:
    """It should set the temperature via the core."""
    subject._api_version = api_version
    mock_task = decoy.mock(cls=Task)
    decoy.when(mock_core.set_target_temperature(42.0)).then_return(mock_task._core)
    result = subject.set_target_temperature(42.0)

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.HEATER_SHAKER_SET_TARGET_TEMPERATURE",
                }
            ),
        ),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )
    decoy.verify(
        mock_core.wait_for_target_temperature(), ignore_extra_args=True, times=0
    )
    if api_version >= APIVersion(2, 27):
        assert isinstance(result, Task)
        assert result._core is mock_task._core
        assert result._api_version == api_version
    else:
        assert result is None


def test_wait_for_temperature(
    decoy: Decoy,
    mock_core: HeaterShakerCore,
    mock_broker: LegacyBroker,
    subject: HeaterShakerContext,
) -> None:
    """It should wait for temperature via the core."""
    subject.wait_for_temperature()

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.HEATER_SHAKER_WAIT_FOR_TEMPERATURE",
                }
            ),
        ),
        mock_core.wait_for_target_temperature(),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )


def test_set_and_wait_for_temperature(
    decoy: Decoy,
    mock_core: HeaterShakerCore,
    mock_broker: LegacyBroker,
    subject: HeaterShakerContext,
) -> None:
    """It should set and wait for the temperature via the core."""
    decoy.when(mock_core.get_target_temperature()).then_return(42.0)

    subject.set_and_wait_for_temperature(celsius=42.0)

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.HEATER_SHAKER_SET_TARGET_TEMPERATURE",
                }
            ),
        ),
        mock_core.set_target_temperature(celsius=42.0),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.HEATER_SHAKER_WAIT_FOR_TEMPERATURE",
                }
            ),
        ),
        mock_core.wait_for_target_temperature(),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )


def test_set_and_wait_for_shake_speed(
    decoy: Decoy,
    mock_core: HeaterShakerCore,
    mock_broker: LegacyBroker,
    subject: HeaterShakerContext,
) -> None:
    """It should set and wait for shake speed via the core."""
    subject.set_and_wait_for_shake_speed(rpm=1337)

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.HEATER_SHAKER_SET_AND_WAIT_FOR_SHAKE_SPEED",
                }
            ),
        ),
        mock_core.set_and_wait_for_shake_speed(rpm=1337),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )


def test_shake_speed(
    decoy: Decoy,
    mock_core: HeaterShakerCore,
    mock_broker: LegacyBroker,
    subject: HeaterShakerContext,
    api_version: APIVersion,
) -> None:
    """It should set shake speed via the core and return a task."""
    subject._api_version = api_version
    mock_task = decoy.mock(cls=Task)
    decoy.when(mock_core.set_shake_speed(1337)).then_return(mock_task._core)
    result = subject.set_shake_speed(1337)
    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.HEATER_SHAKER_SET_SHAKE_SPEED",
                }
            ),
        ),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )
    assert isinstance(result, Task)
    assert result._core is mock_task._core
    assert result._api_version == api_version


def test_open_labware_latch(
    decoy: Decoy,
    mock_core: HeaterShakerCore,
    mock_broker: LegacyBroker,
    subject: HeaterShakerContext,
) -> None:
    """It should open the labware latch via the core."""
    subject.open_labware_latch()

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.HEATER_SHAKER_OPEN_LABWARE_LATCH",
                }
            ),
        ),
        mock_core.open_labware_latch(),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )


def test_close_labware_latch(
    decoy: Decoy,
    mock_core: HeaterShakerCore,
    mock_broker: LegacyBroker,
    subject: HeaterShakerContext,
) -> None:
    """It should close the labware latch via the core."""
    subject.close_labware_latch()

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.HEATER_SHAKER_CLOSE_LABWARE_LATCH",
                }
            ),
        ),
        mock_core.close_labware_latch(),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )


def test_deactivate_shaker(
    decoy: Decoy,
    mock_core: HeaterShakerCore,
    mock_broker: LegacyBroker,
    subject: HeaterShakerContext,
) -> None:
    """It should stop shaking via the core."""
    subject.deactivate_shaker()

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.HEATER_SHAKER_DEACTIVATE_SHAKER",
                }
            ),
        ),
        mock_core.deactivate_shaker(),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )


def test_deactivate_heater(
    decoy: Decoy,
    mock_core: HeaterShakerCore,
    mock_broker: LegacyBroker,
    subject: HeaterShakerContext,
) -> None:
    """It should stop heating via the core."""
    subject.deactivate_heater()

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.HEATER_SHAKER_DEACTIVATE_HEATER",
                }
            ),
        ),
        mock_core.deactivate_heater(),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )


def test_serial_number(
    decoy: Decoy,
    mock_core: HeaterShakerCore,
    subject: HeaterShakerContext,
) -> None:
    """It should get the module's unique serial number."""
    decoy.when(mock_core.get_serial_number()).then_return("abc-123")
    result = subject.serial_number
    assert result == "abc-123"



"""Task handler."""

import asyncio
from datetime import datetime

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.codes import ErrorCodes
from opentrons_shared_data.errors.exceptions import RoboticsInteractionError

from opentrons.protocol_engine.actions import (
    ActionDispatcher,
    FinishTaskAction,
    StartTaskAction,
)
from opentrons.protocol_engine.errors import ErrorOccurrence
from opentrons.protocol_engine.execution.task_handler import TaskHandler
from opentrons.protocol_engine.resources import (
    ModelUtils,
)
from opentrons.protocol_engine.state.state import (
    StateStore,
)
from opentrons.protocol_engine.types import Task


@pytest.fixture
def subject(
    state_store: StateStore,
    model_utils: ModelUtils,
    action_dispatcher: ActionDispatcher,
) -> TaskHandler:
    """Get a task handler to test."""
    return TaskHandler(
        state_store=state_store,
        model_utils=model_utils,
        action_dispatcher=action_dispatcher,
    )


@pytest.fixture
def state_store(decoy: Decoy) -> StateStore:
    """Get a mock in the shape of a StateStore."""
    return decoy.mock(cls=StateStore)


@pytest.fixture
def model_utils(decoy: Decoy) -> ModelUtils:
    """Get a mock in the shape of a StateStore."""
    return decoy.mock(cls=ModelUtils)


@pytest.fixture
def action_dispatcher(decoy: Decoy) -> ActionDispatcher:
    """Get a mock action dispatcher."""
    return decoy.mock(cls=ActionDispatcher)


async def test_create_task(
    subject: TaskHandler,
    decoy: Decoy,
    model_utils: ModelUtils,
    action_dispatcher: ActionDispatcher,
) -> None:
    """Create a task and run it."""
    task_ran = asyncio.Event()

    async def _task(task_handler: TaskHandler) -> None:
        task_ran.set()

    created_timestamp = datetime.now()
    decoy.when(model_utils.get_timestamp()).then_return(created_timestamp)
    decoy.when(model_utils.ensure_id(None)).then_return("create_timestamp")

    task = await subject.create_task(_task)
    await asyncio.wait_for(task_ran.wait(), timeout=0.25)
    await task.asyncioTask
    await asyncio.sleep(0.25)
    assert task.createdAt == created_timestamp
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task=task)),
        action_dispatcher.dispatch(
            FinishTaskAction(
                task_id=matchers.Anything(), finished_at=matchers.Anything(), error=None
            )
        ),
        times=1,
    )


async def test_uses_passed_id(
    subject: TaskHandler,
    decoy: Decoy,
    model_utils: ModelUtils,
    action_dispatcher: ActionDispatcher,
) -> None:
    """Should use provided id."""

    async def _task(task_handler: TaskHandler) -> None:
        await asyncio.sleep(0)

    finished_at = datetime.now()
    decoy.when(model_utils.get_timestamp()).then_return(finished_at)
    decoy.when(model_utils.ensure_id("testid1")).then_return("checked testid1")
    task = await subject.create_task(_task, id="testid1")
    assert task.id == "checked testid1"
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task)),
        times=1,
    )
    await task.asyncioTask
    decoy.verify(
        action_dispatcher.dispatch(
            FinishTaskAction(
                task_id="checked testid1", finished_at=finished_at, error=None
            )
        ),
        times=1,
    )


async def test_generates_id(
    subject: TaskHandler,
    decoy: Decoy,
    model_utils: ModelUtils,
    action_dispatcher: ActionDispatcher,
) -> None:
    """It should generate an id if no id is provided."""

    async def _task(task_handler: TaskHandler) -> None:
        await asyncio.sleep(0)

    finished_at = datetime.now()
    decoy.when(model_utils.get_timestamp()).then_return(finished_at)
    decoy.when(model_utils.ensure_id(None)).then_return("testid2")
    task = await subject.create_task(_task)
    assert task.id == "testid2"
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task)),
        times=1,
    )
    await task.asyncioTask
    decoy.verify(
        action_dispatcher.dispatch(
            FinishTaskAction(task_id="testid2", finished_at=finished_at, error=None)
        ),
        times=1,
    )


async def test_generates_error(
    subject: TaskHandler,
    decoy: Decoy,
    model_utils: ModelUtils,
    action_dispatcher: ActionDispatcher,
) -> None:
    """It should generate an error."""

    async def _task(task_handler: TaskHandler) -> None:
        await asyncio.Event().wait()

    finished_at = datetime.now()
    decoy.when(model_utils.get_timestamp()).then_return(finished_at)
    decoy.when(model_utils.generate_id()).then_return("errorid")
    decoy.when(model_utils.ensure_id(None)).then_return("testid2")
    task = await subject.create_task(_task)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task)),
        times=1,
    )
    task.asyncioTask.cancel(msg="hello")
    try:
        await asyncio.wait_for(task.asyncioTask, timeout=0.25)
    except asyncio.CancelledError:
        pass
    decoy.verify(
        action_dispatcher.dispatch(
            FinishTaskAction(
                task_id="testid2",
                finished_at=finished_at,
                error=ErrorOccurrence.model_construct(
                    id="errorid",
                    createdAt=finished_at,
                    isDefined=False,
                    errorType=matchers.Anything(),
                    errorCode=ErrorCodes.GENERAL_ERROR.value.code,
                    detail=matchers.Anything(),
                    errorInfo=matchers.Anything(),
                    wrappedErrors=matchers.Anything(),
                ),
            )
        ),
        times=1,
    )


async def test_generates_enumerated_error(
    subject: TaskHandler,
    decoy: Decoy,
    model_utils: ModelUtils,
    action_dispatcher: ActionDispatcher,
) -> None:
    """It should generate an enumerated error."""

    async def _task(task_handler: TaskHandler) -> None:
        raise RoboticsInteractionError()

    finished_at = datetime.now()
    decoy.when(model_utils.get_timestamp()).then_return(finished_at)
    decoy.when(model_utils.generate_id()).then_return("errorid")
    decoy.when(model_utils.ensure_id(None)).then_return("testid2")
    task = await subject.create_task(_task)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task)),
        times=1,
    )
    try:
        await asyncio.wait_for(task.asyncioTask, timeout=0.25)
    except (asyncio.CancelledError, RoboticsInteractionError):
        pass
    decoy.verify(
        action_dispatcher.dispatch(
            FinishTaskAction(
                task_id="testid2",
                finished_at=finished_at,
                error=ErrorOccurrence.model_construct(
                    id="errorid",
                    createdAt=finished_at,
                    isDefined=False,
                    errorType=matchers.Anything(),
                    errorCode=ErrorCodes.ROBOTICS_INTERACTION_ERROR.value.code,
                    detail=matchers.Anything(),
                    errorInfo=matchers.Anything(),
                    wrappedErrors=matchers.Anything(),
                ),
            )
        ),
        times=1,
    )


async def test_generates_cancelled_error(
    subject: TaskHandler,
    decoy: Decoy,
    model_utils: ModelUtils,
    action_dispatcher: ActionDispatcher,
) -> None:
    """It should generate an cancelled error."""

    async def _task(task_handler: TaskHandler) -> None:
        await asyncio.Event().wait()

    finished_at = datetime.now()
    decoy.when(model_utils.get_timestamp()).then_return(finished_at)
    decoy.when(model_utils.generate_id()).then_return("errorid")
    decoy.when(model_utils.ensure_id(None)).then_return("testid2")
    task = await subject.create_task(_task)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task)),
        times=1,
    )
    task.asyncioTask.cancel(msg="Cancel task")
    try:
        await asyncio.wait_for(task.asyncioTask, timeout=0.25)
    except asyncio.CancelledError:
        pass
    decoy.verify(
        action_dispatcher.dispatch(
            FinishTaskAction(
                task_id="testid2",
                finished_at=finished_at,
                error=ErrorOccurrence.model_construct(
                    id="errorid",
                    createdAt=finished_at,
                    isDefined=False,
                    errorType=matchers.Anything(),
                    errorCode=ErrorCodes.GENERAL_ERROR.value.code,
                    detail=matchers.StringMatching(r"(CancelledError)|(Cancel Task)"),
                    errorInfo=matchers.Anything(),
                    wrappedErrors=matchers.Anything(),
                ),
            )
        ),
        times=1,
    )


async def test_synchronization_cancel_latest(
    subject: TaskHandler, decoy: Decoy, action_dispatcher: ActionDispatcher
) -> None:
    """Test cancel_lastest synchronization."""
    task1_started = asyncio.Event()
    task2_canceled = asyncio.Event()

    async def task_1_method(task_handler: TaskHandler) -> None:
        """First run method that will proceed."""
        async with task_handler.synchronize_cancel_latest("test"):
            task1_started.set()
            await task2_canceled.wait()

    async def task_2_method(task_handler: TaskHandler) -> None:
        """Second run method that will get canceled."""
        await task1_started.wait()
        try:
            async with task_handler.synchronize_cancel_latest("test"):
                await asyncio.sleep(0)
        except asyncio.CancelledError:
            task2_canceled.set()
            raise

    task1 = await subject.create_task(task_1_method)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task1)),
        times=1,
    )
    task2 = await subject.create_task(task_2_method)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task2)),
        times=1,
    )
    await asyncio.wait_for(
        asyncio.gather(task1.asyncioTask, task2.asyncioTask, return_exceptions=True),
        timeout=1,
    )
    assert task1.asyncioTask.done()
    assert task1.asyncioTask.exception() is None
    assert task2.asyncioTask.done()
    assert task2.asyncioTask.cancelled()


async def test_synchronization_cancel_latest_only_cancels_one_group_id(
    subject: TaskHandler, decoy: Decoy, action_dispatcher: ActionDispatcher
) -> None:
    """Tests cancel_latest synchronization only affects the specified group id."""
    group_id1 = "groupid1"
    group_id2 = "groupid2"

    task1_started = asyncio.Event()
    task2_started = asyncio.Event()
    task3_started = asyncio.Event()

    async def task_1_method(task_handler: TaskHandler) -> None:
        """First run method should proceed (task2 gets cancelled)."""
        async with task_handler.synchronize_cancel_latest(group_id1):
            task1_started.set()
            await asyncio.sleep(0.25)

    async def task_2_method(task_handler: TaskHandler) -> None:
        """Second run method will be canceled (latest)."""
        await task1_started.wait()
        try:
            async with task_handler.synchronize_cancel_latest(group_id1):
                task2_started.set()
                await asyncio.sleep(0.1)
        except asyncio.CancelledError:
            task2_started.set()
            raise

    async def task_3_method(task_handler: TaskHandler) -> None:
        """Third run method (different group) should still be running."""
        async with task_handler.synchronize_cancel_latest(group_id2):
            task3_started.set()
            await asyncio.sleep(0.25)

    task1 = await subject.create_task(task_1_method)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task1)),
        times=1,
    )
    task3 = await subject.create_task(task_3_method)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task3)),
        times=1,
    )
    await task1_started.wait()
    await task3_started.wait()

    task2 = await subject.create_task(task_2_method)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task2)),
        times=1,
    )
    await task2_started.wait()

    assert not task1.asyncioTask.cancelled()
    assert task2.asyncioTask.cancelled()
    assert not task3.asyncioTask.cancelled()
    assert not task3.asyncioTask.done()


async def test_synchronization_cancel_previous(
    subject: TaskHandler, decoy: Decoy, action_dispatcher: ActionDispatcher
) -> None:
    """Test cancel_previous synchronization."""
    task1_started = asyncio.Event()
    task1_canceled = asyncio.Event()

    async def task_1_method(task_handler: TaskHandler) -> None:
        """First run method that will get canceled."""
        async with task_handler.synchronize_cancel_previous("test"):
            task1_started.set()
            await task1_canceled.wait()

    async def task_2_method(task_handler: TaskHandler) -> None:
        """Second run method will finish."""
        await task1_started.wait()
        async with task_handler.synchronize_cancel_previous("test"):
            await asyncio.sleep(0)

    task1 = await subject.create_task(task_1_method)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task1)),
        times=1,
    )
    task2 = await subject.create_task(task_2_method)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task2)),
        times=1,
    )
    await asyncio.wait_for(
        asyncio.gather(task1.asyncioTask, task2.asyncioTask, return_exceptions=True),
        timeout=0.25,
    )
    assert task2.asyncioTask.done()
    assert task2.asyncioTask.exception() is None
    assert task1.asyncioTask.done()
    assert task1.asyncioTask.cancelled()


async def test_synchronization_cancel_previous_only_cancels_one_group_id(
    subject: TaskHandler, decoy: Decoy, action_dispatcher: ActionDispatcher
) -> None:
    """Tests cancel_previous synchronization only affects the specified group id."""
    group_id1 = "groupid1"
    group_id2 = "groupid2"

    task1_started = asyncio.Event()
    task2_started = asyncio.Event()
    task3_started = asyncio.Event()

    async def task_1_method(task_handler: TaskHandler) -> None:
        """First run method that will get canceled."""
        async with task_handler.synchronize_cancel_previous(group_id1):
            task1_started.set()
            try:
                await asyncio.sleep(0)
            except asyncio.CancelledError:
                raise

    async def task_2_method(task_handler: TaskHandler) -> None:
        """Second run method will finish."""
        async with task_handler.synchronize_cancel_previous(group_id1):
            task2_started.set()

    async def task_3_method(task_handler: TaskHandler) -> None:
        """Third run method should still be running."""
        async with task_handler.synchronize_cancel_previous(group_id2):
            task3_started.set()
            await asyncio.sleep(0.25)

    task1 = await subject.create_task(task_1_method)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task1)),
        times=1,
    )
    task3 = await subject.create_task(task_3_method)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task3)),
        times=1,
    )
    await task1_started.wait()
    await task3_started.wait()
    task2 = asyncio.create_task(task_2_method(subject))
    await task2_started.wait()

    assert task1.asyncioTask.cancelled() or task1.asyncioTask.done()
    assert not task2.cancelled()
    assert not task3.asyncioTask.cancelled()
    assert not task3.asyncioTask.done()


async def test_synchronization_sequential(
    subject: TaskHandler, decoy: Decoy, action_dispatcher: ActionDispatcher
) -> None:
    """Test sequential synchronization."""
    task_queue: "asyncio.Queue[str]" = asyncio.Queue()
    task1_started = asyncio.Event()
    task2_started = asyncio.Event()

    async def task_1_method(task_handler: TaskHandler) -> None:
        """First task will finish first."""
        async with task_handler.synchronize_sequential("test"):
            task_queue.put_nowait("task1started")
            task1_started.set()
            task_queue.put_nowait("task1finishedwaiting")
            await task2_started.wait()
            task_queue.put_nowait("task1finished")

    async def task_3_method() -> None:
        task2_started.set()
        task_queue.put_nowait("task2started")

    async def task_2_method(task_handler: TaskHandler) -> None:
        """Second task will finish second."""
        await task1_started.wait()
        synchronizer = task_handler.synchronize_sequential("test")
        await asyncio.gather(synchronizer.__aenter__(), task_3_method())
        await synchronizer.__aexit__(None, None, None)

    task1 = await subject.create_task(task_1_method)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task1)),
        times=1,
    )
    task2 = await subject.create_task(task_2_method)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task2)),
        times=1,
    )
    await asyncio.wait((task1.asyncioTask, task2.asyncioTask), timeout=0.25)
    assert task1.asyncioTask.done()
    assert task2.asyncioTask.done()
    events: list[str] = []
    while not task_queue.empty():
        events.append(task_queue.get_nowait())
    assert events == [
        "task1started",
        "task1finishedwaiting",
        "task2started",
        "task1finished",
    ]


async def test_synchronize_concurrent(
    subject: TaskHandler, decoy: Decoy, action_dispatcher: ActionDispatcher
) -> None:
    """Test concurrent synchronization."""
    task_queue: "asyncio.Queue[str]" = asyncio.Queue()
    task1_started = asyncio.Event()
    task2_started = asyncio.Event()

    async def task_1_method(task_handler: TaskHandler) -> None:
        """Task 1 starts and doesn't finish until task 2 has started."""
        async with task_handler.synchronize_concurrent("test"):
            task_queue.put_nowait("task1started")
            task1_started.set()
            await task2_started.wait()
        task_queue.put_nowait("task1finished")

    async def task_2_method(task_handler: TaskHandler) -> None:
        """Task 2 starts concurrently with task 1."""
        await task1_started.wait()
        async with task_handler.synchronize_concurrent("test"):
            task_queue.put_nowait("task2started")
            task2_started.set()
        task_queue.put_nowait("task2finished")

    task1 = await subject.create_task(task_1_method)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task1)),
        times=1,
    )
    task2 = await subject.create_task(task_2_method)
    decoy.verify(
        action_dispatcher.dispatch(StartTaskAction(task2)),
        times=1,
    )
    await asyncio.wait((task1.asyncioTask, task2.asyncioTask), timeout=0.25)
    assert task1.asyncioTask.done()
    assert task2.asyncioTask.done()
    events: list[str] = []
    while not task_queue.empty():
        events.append(task_queue.get_nowait())
    assert set(events) == {
        "task1started",
        "task2started",
        "task1finished",
        "task2finished",
    }
    assert max(events.index("task1started"), events.index("task2started")) < min(
        events.index("task1finished"), events.index("task2finished")
    )


async def test_cancel_all(
    subject: TaskHandler, decoy: Decoy, state_store: StateStore
) -> None:
    """Test cancel all."""
    mock_tasks = [decoy.mock(cls=Task) for i in range(3)]
    decoy.when(state_store.tasks.get_all_current()).then_return(mock_tasks)
    subject.cancel_all("cancel all")
    for task in mock_tasks:
        decoy.verify(task.asyncioTask.cancel(msg="cancel all"))



"""Camera related state and store resource.

Camera settings, particularly for enablement, can be quieried from via the Camera Provider callback.
However, here Camera settings may also be provided to override or supercede those provided by the callbacks.
"""

from dataclasses import dataclass
from typing import Optional, Tuple

from ..actions import Action
from ._abstract_store import HandlesActions, HasState
from opentrons.protocol_engine.actions import (
    AddCameraCaptureImageSettingsAction,
    AddCameraSettingsAction,
)
from opentrons.protocol_engine.resources.camera_provider import CameraSettings


@dataclass
class CameraState:
    """State of Engine Camera override settings."""

    enablement_settings: Optional[CameraSettings]
    camera_id: Optional[str]
    resolution: Optional[Tuple[int, int]]
    zoom: Optional[float]
    pan: Optional[Tuple[int, int]]
    contrast: Optional[float]
    brightness: Optional[float]
    saturation: Optional[float]


class CameraStore(HasState[CameraState], HandlesActions):
    """Camera container."""

    _state: CameraState

    def __init__(self) -> None:
        """Initialize a Camera store and its state."""
        self._state = CameraState(
            enablement_settings=None,
            camera_id=None,
            resolution=None,
            zoom=None,
            pan=None,
            contrast=None,
            brightness=None,
            saturation=None,
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        if isinstance(action, AddCameraSettingsAction):
            # Update the Camera Enablement settings to the newest override settings
            self._state.enablement_settings = action.enablement_settings
        if isinstance(action, AddCameraCaptureImageSettingsAction):
            # Update the Camera capture image settings to the newest run specific settings
            self._state.camera_id = (
                action.camera_id
                if action.camera_id is not None
                else self._state.camera_id
            )
            self._state.resolution = (
                action.resolution
                if action.resolution is not None
                else self._state.resolution
            )
            self._state.zoom = (
                action.zoom if action.zoom is not None else self._state.zoom
            )
            self._state.pan = action.pan if action.pan is not None else self._state.pan
            self._state.contrast = (
                action.contrast if action.contrast is not None else self._state.contrast
            )
            self._state.brightness = (
                action.brightness
                if action.brightness is not None
                else self._state.brightness
            )
            self._state.saturation = (
                action.saturation
                if action.saturation is not None
                else self._state.saturation
            )


class CameraView:
    """Read-only engine created Camera state view."""

    _state: CameraState

    def __init__(self, state: CameraState) -> None:
        """Initialize the view of Camera state.

        Arguments:
            state: Camera dataclass used for tracking override settings for the camera.
        """
        self._state = state

    def get_enablement_settings(self) -> CameraSettings | None:
        """Get the enablement settings override currently in use. This will take priority over Camera Provider callback provided settings."""
        return self._state.enablement_settings

    def get_camera_id(self) -> str | None:
        """Get the id of the camera set as the default for this run. None means we will use the system default."""
        return self._state.camera_id

    def get_resolution(self) -> Tuple[int, int] | None:
        """Get the resolution set as the default for this run. None means we will use the system default."""
        return self._state.resolution

    def get_zoom(self) -> float | None:
        """Get the resolution set as the default for this run. None means we will use the system default."""
        return self._state.zoom

    def get_pan(self) -> Tuple[int, int] | None:
        """Get the pan values set as the default for this run. None means we will use the system default."""
        return self._state.pan

    def get_contrast(self) -> float | None:
        """Get the contrast set as the default for this run. None means we will use the system default."""
        return self._state.contrast

    def get_brightness(self) -> float | None:
        """Get the brightness set as the default for this run. None means we will use the system default."""
        return self._state.brightness

    def get_saturation(self) -> float | None:
        """Get the saturation set as the default for this run. None means we will use the system default."""
        return self._state.saturation



"""Tests for the CSV Parameter Definitions."""

import inspect

import pytest
from decoy import Decoy

from opentrons.protocol_engine.types import CSVParameter, FileInfo
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.parameters import validation as mock_validation
from opentrons.protocols.parameters.csv_parameter_definition import (
    CSVParameterDefinition,
    create_csv_parameter,
)
from opentrons.protocols.parameters.exceptions import RuntimeParameterRequired


@pytest.fixture(autouse=True)
def _patch_parameter_validation(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    for name, func in inspect.getmembers(mock_validation, inspect.isfunction):
        monkeypatch.setattr(mock_validation, name, decoy.mock(func=func))


@pytest.fixture
def api_version() -> APIVersion:
    """The API version under test."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def csv_parameter_subject(decoy: Decoy) -> CSVParameterDefinition:
    """Return a CSV Parameter Definition subject."""
    decoy.when(mock_validation.ensure_display_name("abc")).then_return("My cool CSV")
    decoy.when(mock_validation.ensure_variable_name("def")).then_return("my_cool_csv")
    decoy.when(mock_validation.ensure_description("ghi")).then_return(
        "Comma Separated Value"
    )
    return create_csv_parameter(
        display_name="abc", variable_name="def", description="ghi"
    )


def test_create_csv_parameter(decoy: Decoy) -> None:
    """It should create a CSV parameter"""
    decoy.when(mock_validation.ensure_display_name("abc")).then_return("My cool CSV")
    decoy.when(mock_validation.ensure_variable_name("def")).then_return("my_cool_csv")
    decoy.when(mock_validation.ensure_description("ghi")).then_return(
        "Comma Separated Value"
    )
    result = create_csv_parameter(
        display_name="abc", variable_name="def", description="ghi"
    )

    assert result._display_name == "My cool CSV"
    assert result.variable_name == "my_cool_csv"
    assert result._description == "Comma Separated Value"
    assert result.value is None
    assert result.file_info is None


def test_set_csv_value(
    decoy: Decoy, csv_parameter_subject: CSVParameterDefinition
) -> None:
    """It should set the CSV parameter value to a byte string."""
    csv_parameter_subject.value = b"123"
    assert csv_parameter_subject.value == b"123"


def test_csv_parameter_as_protocol_engine_type(
    csv_parameter_subject: CSVParameterDefinition,
) -> None:
    """It should return the CSV parameter as a protocol engine type for sending to client."""
    result = csv_parameter_subject.as_protocol_engine_type()
    assert result == CSVParameter(
        displayName="My cool CSV",
        variableName="my_cool_csv",
        description="Comma Separated Value",
        file=None,
    )

    csv_parameter_subject.file_info = FileInfo(id="123", name="abc")
    result = csv_parameter_subject.as_protocol_engine_type()
    assert result == CSVParameter(
        displayName="My cool CSV",
        variableName="my_cool_csv",
        description="Comma Separated Value",
        file=FileInfo(id="123", name="abc"),
    )


def test_csv_parameter_as_csv_parameter_interface(
    api_version: APIVersion,
    csv_parameter_subject: CSVParameterDefinition,
) -> None:
    """It should return the CSV parameter interface for use in a protocol run context."""
    result = csv_parameter_subject.as_csv_parameter_interface(api_version)
    with pytest.raises(RuntimeParameterRequired):
        result.file

    csv_parameter_subject.value = b"abc"
    result = csv_parameter_subject.as_csv_parameter_interface(api_version)
    assert result.contents == "abc"



"""Test engine task core."""

import pytest
from decoy import Decoy, matchers

from opentrons.protocol_engine.clients import SyncClient as EngineClient


@pytest.fixture
def mock_engine_client(decoy: Decoy) -> EngineClient:
    """Get a mock ProtocolEngine synchronous client."""
    return decoy.mock(cls=EngineClient)


def test_get_created_at_timestamp(
    decoy: Decoy, mock_engine_client: EngineClient
) -> None:
    """It should get the created at timestamp from the engine."""
    task_id = "some-id"
    decoy.when(mock_engine_client.state.tasks.get(task_id)).then_return(
        matchers.Anything()
    )


def test_is_done(decoy: Decoy, mock_engine_client: EngineClient) -> None:
    """It should get whether the task is done from the engine."""
    task_id = "some-id"
    decoy.when(mock_engine_client.state.tasks.get_finished(task_id)).then_return(
        matchers.Anything()
    )


def test_is_started(decoy: Decoy, mock_engine_client: EngineClient) -> None:
    """It should get whether the task is started from the engine."""
    task_id = "some-id"
    decoy.when(mock_engine_client.state.tasks.get_current(task_id)).then_return(
        matchers.Anything()
    )


def test_get_finished_at_timestamp(
    decoy: Decoy, mock_engine_client: EngineClient
) -> None:
    """It should get the finished at timestamp from the engine."""
    task_id = "some-id"
    decoy.when(mock_engine_client.state.tasks.get_finished(task_id)).then_return(
        matchers.Anything()
    )



"""Test protocol engine types."""

import pytest
from pydantic import BaseModel, ValidationError

from opentrons.hardware_control.modules.types import (
    AbsorbanceReaderModel,
    FlexStackerModuleModel,
    HeaterShakerModuleModel,
    MagneticBlockModel,
    MagneticModuleModel,
    TemperatureModuleModel,
    ThermocyclerModuleModel,
)
from opentrons.hardware_control.modules.types import (
    ModuleModel as HWModuleModel,
)
from opentrons.protocol_engine.types import (
    HexColor,
    LiquidTrackingType,
    ModuleModel,
    SimulatedProbeResult,
    WellInfoSummary,
)


@pytest.mark.parametrize("hex_color", ["#F00", "#FFCC00CC", "#FC0C", "#98e2d1"])
def test_hex_validation(hex_color: str) -> None:
    """Should allow creating a HexColor."""
    # make sure noting is raised when instantiating this class
    assert HexColor(hex_color)
    assert HexColor.model_validate_json(f'"{hex_color}"')


@pytest.mark.parametrize("invalid_hex_color", ["true", "null", "#123456789"])
def test_handles_invalid_hex(invalid_hex_color: str) -> None:
    """Should raise a validation error."""
    with pytest.raises(ValidationError):
        HexColor(invalid_hex_color)
    with pytest.raises(ValidationError):
        HexColor.model_validate_json(f'"{invalid_hex_color}"')


class _TestModel(BaseModel):
    """Test model for deserializing SimulatedProbeResults."""

    value: LiquidTrackingType


def test_roundtrips_simulated_liquid_probe() -> None:
    """Should be able to roundtrip our simulated results."""
    base = _TestModel(value=SimulatedProbeResult())
    serialized = base.model_dump_json()
    deserialized = _TestModel.model_validate_json(serialized)
    assert isinstance(deserialized.value, SimulatedProbeResult)


def test_roundtrips_nonsimulated_liquid_probe() -> None:
    """Should be able to roundtrip our simulated results."""
    base = _TestModel(value=10.0)
    serialized = base.model_dump_json()
    deserialized = _TestModel.model_validate_json(serialized)
    assert deserialized.value == 10.0


def test_fails_deser_wrong_string() -> None:
    """Should fail to deserialize the wrong string."""
    with pytest.raises(ValidationError):
        _TestModel.model_validate_json('{"value": "not the right string"}')


@pytest.mark.parametrize("height", [None, 10.0, SimulatedProbeResult()])
def test_roundtrips_well_info_summary(height: LiquidTrackingType | None) -> None:
    """It should round trip a WellInfoSummary."""
    inp = WellInfoSummary(
        labware_id="hi",
        well_name="lo",
        loaded_volume=None,
        probed_height=height,
        probed_volume=height,
    )
    outp = WellInfoSummary.model_validate_json(inp.model_dump_json())
    if isinstance(height, SimulatedProbeResult):
        assert outp.labware_id == inp.labware_id
        assert outp.well_name == inp.well_name
        assert isinstance(outp.probed_height, SimulatedProbeResult)
        assert isinstance(outp.probed_volume, SimulatedProbeResult)
    else:
        assert outp == inp


@pytest.mark.parametrize("op_1", [SimulatedProbeResult(), 100.0, -5])
@pytest.mark.parametrize("op_2", [SimulatedProbeResult(), 100.0, -5])
def test_simulated_probe_result_operand_math(
    op_1: LiquidTrackingType, op_2: LiquidTrackingType
) -> None:
    """Ensure that math operators can be used with SimulatedProbeResult."""
    _error = None
    r: LiquidTrackingType | None = None
    try:
        r = op_1 + op_2
        r = op_1 - op_2
        r = op_1 / op_2
        r = op_1 * op_2
        r = op_1**op_2
        r = op_1 // op_2
        r = op_1 % op_2
        r = op_1 < op_2
        r = op_1 > op_2
        r = op_1 >= op_2
        r = op_1 <= op_2
        r = op_1 == op_2
    except Exception as _e:
        _error = _e
    assert _error is None
    assert r is not None


@pytest.mark.parametrize(
    "hardware_module_model",
    [m for m in MagneticModuleModel]
    + [m for m in TemperatureModuleModel]
    + [m for m in ThermocyclerModuleModel]
    + [m for m in HeaterShakerModuleModel]
    + [m for m in MagneticBlockModel]
    + [m for m in AbsorbanceReaderModel]
    + [m for m in FlexStackerModuleModel],
)
def test_module_model_translation(hardware_module_model: HWModuleModel) -> None:
    """It should turn every hardware module model into an engine model."""
    engine_module = ModuleModel.from_hardware(hardware_module_model)
    assert engine_module.value == hardware_module_model.value
